@@ -16,5 +16,5 @@
 from ._client import OAuth2Client  # noqa: F401
 from ._exception import OAuth2ExpiredTokenException, OAuth2RevokedTokenException, \
     OAuth2TokenException, RESTException  # noqa: F401
-from ._type import CommentsSort, ListingSort, Me, OAuth2Scope, OAuth2Scopes, OAuth2Token, \
-    RateLimit, Submission, SubmitKind, Subreddit  # noqa: F401
+from ._type import CommentsSort, LinkThing, Listing, ListingSort, Me, OAuth2Scope, OAuth2Scopes, \
+    OAuth2Token, RateLimit, Submission, SubmitKind, Subreddit  # noqa: F401
